@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 
 from .trainer import NerTrainer, load_model, load_pretrained_model, load_pretrained_tokenizer
-from .dataset import InputExample, encode_examples, examples_to_dataset
+from .dataset import InputExample, encode_examples
 from .dataset import to_BIOS, load_examples_from_bios_file, export_bios_file
 from .utils import get_entities
 from .args import get_args
