@@ -1,13 +1,15 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
+import os
 import numpy as np
 from loguru import logger
-import mlflow
+#  import mlflow
 
 import torch
 import torch.nn as nn
-from torch.nn import CrossEntropyLoss
+from torch.nn import CrossEntropyLoss, MSELoss, BCEWithLogitsLoss
+from ...losses import FocalLoss, DiceLoss
 import torch.nn.functional as F
 #  from ..models.crf_0 import CRF as CRF0
 #  from ..models.crf import CRF
@@ -30,7 +32,8 @@ from ...utils.multiprocesses import barrier_leader_process, barrier_member_proce
 
 from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
 from transformers import AutoConfig, BertTokenizerFast, AutoModelForTokenClassification
-from transformers.modeling_bert import BertPreTrainedModel, BertModel
+#  from transformers.modeling_bert import BertPreTrainedModel, BertModel
+from transformers.models.bert import BertModel, BertPreTrainedModel
 
 CRF_TYPE = "old_crf"
 CRF_TYPE = "new_crf"
@@ -165,6 +168,10 @@ class BertCrfForNer(BertPreTrainedModel):
         self.num_labels = config.num_labels
         self.crf_type = config.crf_type
         self.no_crf_loss = config.no_crf_loss
+        self.loss_type = config.loss_type
+        self.focalloss_gamma = config.focalloss_gamma
+        self.focalloss_alpha = config.focalloss_alpha
+        self.diceloss_weight = config.diceloss_weight
 
         self.bert = BertModel(config)
         self.dropout = nn.Dropout(config.hidden_dropout_prob)
@@ -245,9 +252,42 @@ class BertCrfForNer(BertPreTrainedModel):
         logits = self.classifier(sequence_output)
         outputs = (logits, )
         if labels is not None:
-            loss_fct = CrossEntropyLoss()
+
+            #  loss_fct = CrossEntropyLoss()
+
+            assert self.loss_type in [
+                'FocalLoss', 'CrossEntropyLoss', 'BCEWithLogitsLoss'
+            ]
+            if self.loss_type == 'FocalLoss':
+                loss_fct = FocalLoss(gamma=self.focalloss_gamma,
+                                     alpha=self.focalloss_alpha)
+
+                #  loss = FocalLoss(gamma=self.focalloss_gamma,
+                #                   alpha=self.focalloss_alpha)(logits.view(
+                #                       -1, self.num_labels), labels.view(-1))
+            elif self.loss_type == 'DiceLoss':
+                loss_fct = DiceLoss(weight=self.diceloss_weight)
+
+            elif self.loss_type == 'BCEWithLogitsLoss':
+                loss_fct = BCEWithLogitsLoss()
+                #  logger.debug(f"logits: {logits.shape}, {logits}")
+                #  logger.debug(f"labels: {labels.shape}, {labels}")
+                #  loss = loss_fct(logits, labels.float())
+                #  loss = loss_fct(logits.view(-1, self.num_labels),
+                #                  labels.view(-1, self.num_labels).float())
+            elif self.loss_type == 'CrossEntropyLoss':
+                loss_fct = CrossEntropyLoss()
+                #  logger.warning(f"logits: {logits.shape}")
+                #  logger.warning(f"labels: {labels.shape}")
+                #  loss = loss_fct(logits.view(-1, self.num_labels),
+                #                  labels.view(-1))
+            else:
+                raise Exception(
+                    f"Bad loss_type: {self.loss_type}. Must be one of 'FocalLoss', 'CrossEntropyLoss', 'BCEWithLogitsLoss'."
+                )
+
             # Only keep active parts of the loss
-            if attention_mask is not None:
+            if self.loss_type == 'CrossEntropyLoss' and attention_mask is not None:
                 active_loss = attention_mask.view(-1) == 1
                 active_logits = logits.view(-1, self.num_labels)
                 active_labels = torch.where(
@@ -401,12 +441,17 @@ MODEL_CLASSES = {
 
 
 def load_pretrained_tokenizer(args):
-    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
-    tokenizer = tokenizer_class.from_pretrained(
-        args.model_path,
-        do_lower_case=args.do_lower_case,
-        cache_dir=args.cache_dir if args.cache_dir else None,
-    )
+    #  config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
+    #  tokenizer = tokenizer_class.from_pretrained(
+    #      args.model_path,
+    #      do_lower_case=args.do_lower_case,
+    #      is_english=args.is_english,
+    #      cache_dir=args.cache_dir if args.cache_dir else None,
+    #  )
+    from ..token_utils import HFTokenizer
+    tokenizer = HFTokenizer(os.path.join(args.model_path, 'vocab.txt'),
+                            lowercase=args.do_lower_case,
+                            cc=args.cc)
 
     return tokenizer
 
@@ -425,6 +470,10 @@ def load_pretrained_model(args):
     )
     setattr(config, 'crf_type', args.crf_type)
     setattr(config, 'no_crf_loss', args.no_crf_loss)
+    setattr(config, 'loss_type', args.loss_type)
+    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
+    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
+    setattr(config, 'diceloss_weight', args.diceloss_weight)
 
     logger.info(f"model_path: {args.model_path}")
     logger.info(f"config:{config}")
@@ -458,14 +507,22 @@ def collate_fn(batch):
     batch should be a list of (sequence, target, length) tuples...
     Returns a padded tensor of sequences sorted from longest to shortest,
     """
-    all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens = map(
-        torch.stack, zip(*batch))
-    max_len = max(all_lens).item()
+    #  all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens = map(
+    #      torch.stack, zip(*batch))
+    all_input_ids = torch.stack([e.input_ids for e in batch])
+    all_attention_mask = torch.stack([e.attention_mask for e in batch])
+    all_token_type_ids = torch.stack([e.token_type_ids for e in batch])
+    all_labels = torch.stack([e.labels for e in batch])
+    all_input_lens = torch.stack([e.input_len for e in batch])
+    all_token_offsets = torch.stack([e.token_offsets for e in batch])
+
+    max_len = max(all_input_lens).item()
     all_input_ids = all_input_ids[:, :max_len]
     all_attention_mask = all_attention_mask[:, :max_len]
     all_token_type_ids = all_token_type_ids[:, :max_len]
     all_labels = all_labels[:, :max_len]
-    return all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens
+
+    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_labels
 
 
 def load_model(args):
@@ -490,6 +547,7 @@ def build_default_model(args):
         model, args.weight_decay)
     optimizer = AdamW(optimizer_parameters,
                       lr=args.learning_rate,
+                      eps=args.adam_epsilon,
                       correct_bias=False)
 
     # -------- scheduler --------
@@ -503,13 +561,24 @@ def build_default_model(args):
     return model, optimizer, scheduler
 
 
-def preds_to_entities(preds, id2label, markup='bios', autofix=False):
+def preds_to_entities(preds,
+                      id2label,
+                      token_offsets,
+                      markup='bios',
+                      autofix=False):
     #  logger.debug(f"{preds}")
     preds = [id2label[x] for x in preds]
     label_entities = get_entities(preds,
                                   id2label=id2label,
                                   markup=markup,
                                   autofix=autofix)
+
+    label_entities = [(c, token_offsets[s + 1][0],
+                       token_offsets[e + 1][-1] - 1)
+                      for c, s, e in label_entities]
+    label_entities = [
+        x for x in label_entities if x[1] <= x[2] and x[1] >= 0 and x[2] >= 0
+    ]
     return label_entities
 
 
@@ -547,8 +616,8 @@ class NerTrainer(Trainer):
         inputs = {
             "input_ids": batch[0],
             "attention_mask": batch[1],
-            "labels": batch[3] if known_labels else None,
-            'input_lens': batch[4]
+            "labels": batch[5] if known_labels else None,
+            'input_lens': batch[3]
         }
         if args.model_type != "distilbert":
             # XLM and RoBERTa don"t use segment_ids
@@ -556,10 +625,15 @@ class NerTrainer(Trainer):
                 batch[2] if args.model_type in ["bert", "xlnet"] else None)
         return inputs
 
-    def examples_to_dataset(self, examples, max_seq_length):
-        from .dataset import examples_to_dataset
-        return examples_to_dataset(examples, self.label2id, self.tokenizer,
-                                   max_seq_length)
+    #  def examples_to_dataset(self, examples, max_seq_length):
+    #      from .dataset import examples_to_dataset
+    #      return examples_to_dataset(examples, self.label2id, self.tokenizer,
+    #                                 max_seq_length)
+
+    def encode_examples(self, examples, max_seq_length):
+        from .dataset import encode_examples
+        return encode_examples(examples, self.label2id, self.tokenizer,
+                               max_seq_length)
 
     #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
     #
@@ -579,7 +653,7 @@ class NerTrainer(Trainer):
         self.error_entities = []
 
     #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
-    def on_eval_step(self, args, model, step, batch, batch_features):
+    def on_eval_step(self, args, model, step, batch):
         #  outputs = self.on_train_step(args, model, step, batch)
 
         inputs = self.batch_to_inputs(args, batch)
@@ -700,11 +774,11 @@ class NerTrainer(Trainer):
         eval_info, _ = self.metric.result()
         results = {f'{key}': value for key, value in eval_info.items()}
 
-        if args.do_experiment:
-            #  logger.debug(f"loss: {loss}")
-            #  mlflow.log_metric('loss', loss.item())
-            for key, value in eval_info.items():
-                mlflow.log_metric(key, value)
+        #  if args.do_experiment:
+        #      #  logger.debug(f"loss: {loss}")
+        #      #  mlflow.log_metric('loss', loss.item())
+        #      for key, value in eval_info.items():
+        #          mlflow.log_metric(key, value)
 
         #  batch_preds = tags
         #  for i in range(len(batch_preds)):
@@ -722,6 +796,9 @@ class NerTrainer(Trainer):
     #  def on_predict_step(self, args, test_dataset, step, model, inputs,
     #                      outputs):
     def on_predict_step(self, args, model, step, batch):
+
+        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_labels = batch
+
         inputs = self.batch_to_inputs(args, batch, known_labels=False)
         attention_mask = inputs['attention_mask']
 
@@ -788,7 +865,7 @@ class NerTrainer(Trainer):
 
         #  batch_preds = model.crf.decode(logits)
 
-        def to_entities(args, preds):
+        def to_entities(args, preds, token_offsets):
             #  logger.debug(f"{preds}")
 
             #  preds = [args.id2label[x] for x in preds]
@@ -796,7 +873,9 @@ class NerTrainer(Trainer):
             #                                args.id2label,
             #                                args.markup,
             #                                autofix=args.autofix)
-            label_entities = preds_to_entities(preds, id2label=args.id2label)
+            label_entities = preds_to_entities(preds,
+                                               id2label=args.id2label,
+                                               token_offsets=token_offsets)
 
             #  logger.debug(f"{label_entities}")
             json_d = {}
@@ -813,9 +892,10 @@ class NerTrainer(Trainer):
         #  json_d = to_entities(args, preds)
         #  self.pred_results.append(json_d)
 
-        for preds in batch_preds:
+        for i, preds in enumerate(batch_preds):
+            token_offsets = all_token_offsets[i]
             preds = preds[1:-1]  # [CLS]XXXX[SEP]
-            json_d = to_entities(args, preds)
+            json_d = to_entities(args, preds, token_offsets)
             self.pred_results.append(json_d)
 
     def on_predict_end(self, args, test_dataset):
