@@ -1,24 +1,30 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-import os, json, copy
+import copy, random
+import json
+import os
 from pathlib import Path
-from loguru import logger
-import numpy as np
+from collections import defaultdict
 
+import numpy as np
 import torch
+from loguru import logger
 from torch.autograd import Variable
 from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
+from tqdm import tqdm
+
 from ..utils import init_theta
+from ..utils.multiprocesses import (barrier_leader_process,
+                                    barrier_member_processes,
+                                    is_master_process, is_multi_processes)
+from ..utils.progbar import Progbar
 
 try:
     from torch.utils.tensorboard import SummaryWriter
 except ImportError:
     from tensorboardX import SummaryWriter
 
-from ..utils.progbar import Progbar
-from ..utils.multiprocesses import is_master_process, is_multi_processes, barrier_leader_process, barrier_member_processes
-
 
 def get_default_optimizer_parameters(model, weight_decay):
     param_optimizers = list(model.named_parameters())
@@ -40,6 +46,49 @@ def get_default_optimizer_parameters(model, weight_decay):
     return optimizer_grouped_parameters
 
 
+enttype_labels = defaultdict(list)
+
+
+def augment_examples(examples, augs=0):
+    for a in range(augs):
+        aug_examples = copy.deepcopy(examples)
+        for e in tqdm(aug_examples, desc="Aug examples"):
+            p0 = 0
+            text = copy.deepcopy(e.text_a)
+            new_text = ""
+            new_labels = []
+            for i, (label, start, end) in enumerate(e.labels):
+                new_text += text[p0:start]
+                if random.randint(0, 1000 - 1) < 666:
+                    new_start = len(new_text)
+                    new_mention = text[start:end + 1]
+                    new_text += new_mention
+                    new_labels.append(
+                        (label, new_start, new_start + len(new_mention) - 1))
+                    p0 = end + 1
+                else:
+                    new_start = len(new_text)
+                    new_idx = random.randint(0, len(enttype_labels[label]) - 1)
+                    new_mention = enttype_labels[label][new_idx]
+                    new_text += new_mention
+                    new_end = new_start + len(new_mention) - 1
+                    new_labels.append((label, new_start, new_end))
+                    p0 = end + 1
+            new_text += text[p0:]
+            #  logger.debug(
+            #      f"text: {text} | {[ text[start:end+1] for _, start, end in e.labels]}"
+            #  )
+            #  logger.warning(
+            #      f"new_text: {new_text} | {[ new_text[start:end+1] for _, start, end in new_labels]}"
+            #  )
+            e.text_a = new_text
+            e.labels = new_labels
+
+        examples.extend(aug_examples)
+
+    return examples
+
+
 def generate_dataloader(args,
                         dataset,
                         batch_size,
@@ -49,13 +98,84 @@ def generate_dataloader(args,
     Sampler = SequentialSampler if keep_order else RandomSampler
     sampler = DistributedSampler(dataset) if is_multi_processes(
         args) else Sampler(dataset)
-    dataloader = DataLoader(dataset,
-                            sampler=sampler,
-                            batch_size=batch_size,
-                            collate_fn=collate_fn)
+    dataloader = DataLoader(
+        dataset,
+        sampler=sampler,
+        batch_size=batch_size,
+        #  collate_fn=None)
+        collate_fn=collate_fn)
     return dataloader
 
 
+def common_batch_encode(texts, label2id, tokenizer, max_seq_length):
+    all_encodes = tokenizer.batch_encode(texts, add_special_tokens=True)
+    all_tokens = all_encodes['tokens']
+    all_token_offsets = all_encodes['offsets']
+    all_token2char = all_encodes['token2char']
+    all_char2token = all_encodes['char2token']
+
+    all_input_ids = all_encodes['ids']
+    all_attention_mask = all_encodes['attention_mask']
+    all_token_type_ids = all_encodes['type_ids']
+
+    all_input_lens = [len(tokens) for tokens in all_tokens]
+
+    all_padding_lens = [max_seq_length - n for n in all_input_lens]
+    for i, (input_ids, attention_mask, token_type_ids, token2char,
+            token_offsets, padding_length) in enumerate(
+                tqdm(zip(all_input_ids, all_attention_mask, all_token_type_ids,
+                         all_token2char, all_token_offsets, all_padding_lens),
+                     desc="common_batch_encode")):
+
+        if padding_length < 0:
+            all_tokens[i] = all_tokens[i][:max_seq_length - 1] + ['[SEP]']
+            all_input_ids[i] = input_ids[:max_seq_length - 1] + [102]
+            all_attention_mask[i] = attention_mask[:max_seq_length]
+            all_token_type_ids[i] = token_type_ids[:max_seq_length]
+            all_token2char[i] = token2char[:max_seq_length]
+            all_token_offsets[i] = token_offsets[:max_seq_length - 1] + [(0, 0)
+                                                                         ]
+            #  logger.warning(f"all_tokens[i]: {all_tokens[i]}")
+            #  logger.debug(f"all_input_ids[i]: {all_input_ids[i]}")
+            #  logger.debug(f"all_token_offsets[i]: {all_token_offsets[i]}")
+
+        else:
+            all_input_ids[i] = input_ids + [0] * padding_length
+            all_attention_mask[i] = attention_mask + [0] * padding_length
+            all_token_type_ids[i] = token_type_ids + [0] * padding_length
+            all_token2char[i] = token2char + [0] * padding_length
+            all_token_offsets[i] = token_offsets + [(0, 0)] * padding_length
+        all_input_lens[i] = len(input_ids)
+
+    #  logger.warning(f"all_input_ids: {all_input_ids}")
+    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
+    all_attention_mask = torch.from_numpy(
+        np.array(all_attention_mask, dtype=np.int64))
+    all_token_type_ids = torch.from_numpy(
+        np.array(all_token_type_ids, dtype=np.int64))
+    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
+    all_token_offsets = torch.from_numpy(
+        np.array(all_token_offsets, dtype=np.int64))
+
+    return (all_tokens, all_token2char, all_char2token, all_input_ids,
+            all_attention_mask, all_token_type_ids, all_input_lens,
+            all_token_offsets)
+
+
+def common_to_tensors(all_input_ids, all_attention_mask, all_token_type_ids,
+                      all_input_lens, all_token_offsets):
+    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
+    all_attention_mask = torch.from_numpy(
+        np.array(all_attention_mask, dtype=np.int64))
+    all_token_type_ids = torch.from_numpy(
+        np.array(all_token_type_ids, dtype=np.int64))
+    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
+    all_token_offsets = torch.from_numpy(
+        np.array(all_token_offsets, dtype=np.int64))
+
+    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets
+
+
 class Trainer:
     def __init__(self, args):
         init_theta(args)
@@ -63,13 +183,22 @@ class Trainer:
         self.args = args
         self.collate_fn = None
 
+        self.eval_loss = 0.0
+
     #  def build_model(self, args):
     #      raise NotImplementedError
 
     def batch_to_inputs(self, args, batch, known_labels=True):
         raise NotImplementedError
 
-    def examples_to_dataset(self, examples, max_seq_length):
+    #  def examples_to_dataset(self, examples, max_seq_length):
+    #      raise NotImplementedError
+
+    def encode_examples(self,
+                        examples,
+                        max_seq_length,
+                        enttype_labels=None,
+                        epoch=-1):
         raise NotImplementedError
 
     #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
@@ -77,6 +206,7 @@ class Trainer:
     #      return generate_dataloader(args, dataset, batch_size, keep_order)
 
     def on_train_step(self, args, model, step, batch):
+
         inputs = self.batch_to_inputs(args, batch)
         outputs = model(**inputs)
         return outputs
@@ -88,12 +218,13 @@ class Trainer:
     #      results = {}
     #      return results
 
-    def on_eval_step(self, args, model, step, batch, batch_features):
+    #  def on_eval_step(self, args, model, step, batch, batch_features):
+    def on_eval_step(self, args, model, step, batch):
         inputs = self.batch_to_inputs(args, batch)
         outputs = model(**inputs)
         return outputs, {}
 
-    def on_eval_end(self, args, eval_dataset):
+    def on_eval_end(self, args, eval_examples):
         results = {}
         return results
 
@@ -109,16 +240,15 @@ class Trainer:
         return []
 
     def save_args(self, args, model_path):
-        #  torch.save(args, os.path.join(model_path, "training_args.bin"))
-        json.dump(
-            {
-                k: v
-                for k, v in args.__dict__.items() if v is None
-                or type(v) in [bool, str, int, float, dict, list, tuple]
-            },
-            open(os.path.join(model_path, "training_args.json"), 'w'),
-            ensure_ascii=False,
-            indent=2)
+        from .utils import save_args
+        save_args(args, model_path)
+
+    def save_params_file(self, args, model_path):
+        latest_dir = args.latest_dir
+        dataset_name = args.dataset_name
+        cmd_cp_params = f"cp {dataset_name}_params.py {model_path}/{dataset_name}_params.py"
+        logger.warning(f"{cmd_cp_params}")
+        os.system(cmd_cp_params)
 
     def save_model(self, args, model, tokenizer, optimizer, scheduler,
                    model_path):
@@ -134,10 +264,10 @@ class Trainer:
 
         self.save_args(args, model_path)
 
-        torch.save(optimizer.state_dict(),
-                   os.path.join(model_path, "optimizer.pt"))
-        torch.save(scheduler.state_dict(),
-                   os.path.join(model_path, "scheduler.pt"))
+        #  torch.save(optimizer.state_dict(),
+        #             os.path.join(model_path, "optimizer.pt"))
+        #  torch.save(scheduler.state_dict(),
+        #             os.path.join(model_path, "scheduler.pt"))
 
     def train(self, args, train_examples, eval_examples):
 
@@ -146,15 +276,34 @@ class Trainer:
         )
 
         self.save_args(args, args.latest_dir)
+        self.save_params_file(args, args.latest_dir)
 
         #  if is_master_process(args):
         #      tb_writer = SummaryWriter()
 
-        train_dataset, _ = self.examples_to_dataset(train_examples,
-                                                    args.train_max_seq_length)
+        #  _, all_features = self.examples_to_dataset(train_examples,
+        #                                             args.train_max_seq_length)
+
+        for e in tqdm(train_examples + eval_examples,
+                      desc="Build enttype_labels"):
+            text = e.text_a
+            for label, start, end in e.labels:
+                mention = text[start:end + 1]
+                enttype_labels[label].append(mention)
+        for k, labels in enttype_labels.items():
+            enttype_labels[k] = list(set(enttype_labels[k]))
+
+        # FIXME
+        train_examples = augment_examples(train_examples,
+                                          augs=args.num_augments)
+        all_features = self.encode_examples(train_examples,
+                                            args.train_max_seq_length)
+
+        logger.warning(f"Generate dataloader...")
         train_dataloader = generate_dataloader(
             args,
-            train_dataset,
+            #  train_dataset,
+            all_features,
             batch_size=args.per_gpu_train_batch_size,
             keep_order=False,
             collate_fn=self.collate_fn)
@@ -183,13 +332,16 @@ class Trainer:
         tokenizer = self.tokenizer
 
         # Check if saved optimizer or scheduler states exist
-        model_path = Path(args.model_path)
-        optimizer_saved_file = model_path / "optimizer.pt"
-        scheduler_saved_file = model_path / "scheduler.pt"
-        if optimizer_saved_file.exists() and scheduler_saved_file.exists():
-            optimizer.load_state_dict(torch.load(optimizer_saved_file))
-            scheduler.load_state_dict(torch.load(scheduler_saved_file))
+        #  model_path = Path(args.model_path)
+        #  optimizer_saved_file = model_path / "optimizer.pt"
+        #  scheduler_saved_file = model_path / "scheduler.pt"
+        #  if optimizer_saved_file.exists() and scheduler_saved_file.exists():
+        #      optimizer.load_state_dict(torch.load(optimizer_saved_file))
+        #      scheduler.load_state_dict(torch.load(scheduler_saved_file))
 
+        model = model.to(args.device)
+        #  logger.warning(
+        #      f"model.named_parameters(): {list(model.named_parameters())}")
         if args.fp16:
             try:
                 from apex import amp
@@ -220,23 +372,23 @@ class Trainer:
         model_path = Path(args.model_path)
         output_dir = Path(args.output_dir)
 
-        if model_path.exists() and "checkpoint" in str(model_path):
-            # set trained_steps to trained_steps of last saved checkpoint from model path
-            trained_steps = int(model_path.parts()[-1].split("-")[-1])
-            epochs_trained = trained_steps // (
-                len(train_dataloader) // args.gradient_accumulation_steps)
-            steps_trained_in_current_epoch = trained_steps % (
-                len(train_dataloader) // args.gradient_accumulation_steps)
-
-            logger.info(
-                "  Continuing training from checkpoint, will skip to saved trained_steps"
-            )
-            logger.info("  Continuing training from epoch %d", epochs_trained)
-            logger.info("  Continuing training from global step %d",
-                        trained_steps)
-            logger.info("  Will skip the first %d steps in the first epoch",
-                        steps_trained_in_current_epoch)
-
+        #  if model_path.exists() and "checkpoint" in str(model_path):
+        #      # set trained_steps to trained_steps of last saved checkpoint from model path
+        #      trained_steps = int(model_path.parts()[-1].split("-")[-1])
+        #      epochs_trained = trained_steps // (
+        #          len(train_examples) // args.gradient_accumulation_steps)
+        #      steps_trained_in_current_epoch = trained_steps % (
+        #          len(train_examples) // args.gradient_accumulation_steps)
+        #
+        #      logger.info(
+        #          "  Continuing training from checkpoint, will skip to saved trained_steps"
+        #      )
+        #      logger.info("  Continuing training from epoch %d", epochs_trained)
+        #      logger.info("  Continuing training from global step %d",
+        #                  trained_steps)
+        #      logger.info("  Will skip the first %d steps in the first epoch",
+        #                  steps_trained_in_current_epoch)
+        #
         train_loss, logging_loss = 0.0, 0.0
 
         best_index = args.best_index
@@ -265,7 +417,16 @@ class Trainer:
             sda_loss_fct = MSELoss()
             history_logits = []
             sda_teachers = args.sda_teachers
-            teacher_models = []
+            sda_stategy = args.sda_stategy
+
+            if args.sda_empty_first:
+                teacher_models = []
+            else:
+                t_model = copy.deepcopy(model)
+                t_model.eval()
+                teacher_models = [t_model]
+
+            #  best_logits = []
 
         #  train_iterator = trange(
         #      epochs_trained,
@@ -275,12 +436,27 @@ class Trainer:
         #  )
         #  for epoch in train_iterator:
         for epoch in range(epochs_trained, args.num_train_epochs):
+            args.epoch = epoch
 
-            #  epoch_iterator = tqdm(train_dataloader,
-            #                        desc="Iteration",
-            #                        disable=args.local_rank not in [-1, 0])
-            #  for step, batch in enumerate(epoch_iterator):
-
+            #  all_features = self.encode_examples(train_examples,
+            #                                      args.train_max_seq_length,
+            #                                      enttype_labels=enttype_labels,
+            #                                      epoch=epoch)
+            #
+            #  logger.warning(f"Generate dataloader...")
+            #  train_dataloader = generate_dataloader(
+            #      args,
+            #      #  train_dataset,
+            #      all_features,
+            #      batch_size=args.per_gpu_train_batch_size,
+            #      keep_order=False,
+            #      collate_fn=self.collate_fn)
+            #  logger.info(f"Start training ...")
+            #  logger.info(f"  Num examples    = {len(train_examples)}")
+            #  logger.info(f"  Num epoch steps = {len(train_dataloader)}")
+            #  logger.info(f"  Num epochs = {args.num_train_epochs}")
+            #  logger.info(f"  Batch size = {args.per_gpu_train_batch_size}")
+            #
             pbar = Progbar(target=len(train_dataloader),
                            stateful_metrics=['loss'],
                            desc=f"Epoch({epoch+1}/{args.num_train_epochs})")
@@ -290,7 +466,9 @@ class Trainer:
                     continue
 
                 model.train()
-                batch = tuple(t.to(args.device) for t in batch)
+                batch = tuple(
+                    t.to(args.device) if isinstance(t, torch.Tensor) else t
+                    for t in batch)
 
                 outputs = self.on_train_step(args, model, step, batch)
                 #  inputs = self.batch_to_inputs(args, batch)
@@ -300,6 +478,8 @@ class Trainer:
                 # -------- loss --------
                 loss, logits = outputs[:2]
 
+                #  loss += torch.tensor(self.eval_loss).cuda()
+
                 if enable_kd:
                     inputs = self.batch_to_inputs(args, batch)
                     if "labels" in inputs:
@@ -325,6 +505,14 @@ class Trainer:
                         #  sda_loss = Variable(sda_loss, requires_grad=True)
                         loss += sda_loss * args.sda_coeff
 
+                    #  if best_logits:
+                    #      sda_logits = torch.stack(best_logits)
+                    #      sda_logits = torch.mean(sda_logits, dim=0)
+                    #      if sda_logits.shape[0] == logits.shape[0]:
+                    #          sda_loss = sda_loss_fct(logits, sda_logits)
+                    #          #  sda_loss = Variable(sda_loss, requires_grad=True)
+                    #          loss += sda_loss * args.sda_coeff
+
                 #  loss = Variable(loss, requires_grad=True)
                 #  inputs = self.batch_to_inputs(args, batch)
                 #  logger.debug(f"inputs: {inputs}")
@@ -337,6 +525,7 @@ class Trainer:
                     loss = loss.mean()
                 if args.gradient_accumulation_steps > 1:
                     loss = loss / args.gradient_accumulation_steps
+                #  logger.info(f"loss: {loss}")
 
                 if args.fp16:
                     with amp.scale_loss(loss, optimizer) as scaled_loss:
@@ -347,7 +536,8 @@ class Trainer:
                 lr = scheduler.get_last_lr()[0]
                 #  lr = scheduler.get_lr()[0]
                 pbar.update(step + 1,
-                            values=[('lr', lr), ('loss', loss.item())])
+                            values=[('lr', lr), ('loss', loss.item()),
+                                    ('eval_loss', self.eval_loss)])
                 train_loss += loss.item()
 
                 if (step + 1) % args.gradient_accumulation_steps == 0:
@@ -438,6 +628,9 @@ class Trainer:
                         #  best_index = 'f1'  # ['f1', 'acc', 'recall', 'loss']
                         best_index = args.best_index
                         eval_value = eval_results[best_index]
+                        #  logger.warning(
+                        #      f"best_index: {best_index}, best_value: {best_value:.6f}, eval_value: {eval_value:.6f}"
+                        #  )
                         is_best = False
                         if best_index in ['loss']:
                             if eval_value < best_value:
@@ -447,8 +640,13 @@ class Trainer:
                                 is_best = True
                         if is_best:
                             logger.warning(
-                                f"Best {best_index}: {eval_value:.4f} ({eval_value - best_value:.6f})"
+                                f"Best {best_index}: {eval_value:.6f} ({eval_value - best_value:.6f})"
                             )
+                            #  if enable_sda:
+                            #      best_logits.append(logits.detach())
+                            #      if len(best_logits) > sda_teachers:
+                            #          best_logits = best_logits[1:]
+
                             best_value = eval_value
 
                             #  bestmodel_path = output_dir / f"best_fold{args.fold}"
@@ -463,14 +661,49 @@ class Trainer:
                                     best_symlink.unlink()
                                 os.symlink(f"../{checkpoint_dir}",
                                            best_symlink)
+                        else:
+                            logger.info(
+                                f"dev-{best_index}/best-{best_index}: {eval_value:.6f}/{best_value:.6f}"
+                            )
+
+            #  if enable_sda:
+            #      best_logits.append(logits.detach())
+            #      if len(best_logits) > sda_teachers:
+            #          best_logits = best_logits[1:]
 
             if enable_sda:
-                t_model = copy.deepcopy(model)
-                t_model.eval()
-                teacher_models.append(t_model)
-                assert len(teacher_models) <= sda_teachers + 1
-                if len(teacher_models) == sda_teachers + 1:
-                    teacher_models = teacher_models[1:]
+                if sda_stategy == "recent_models":
+                    if len(teacher_models) >= sda_teachers:
+                        teacher_models = teacher_models[1:sda_teachers + 1]
+                    t_model = copy.deepcopy(model)
+                    t_model.eval()
+                    teacher_models.append(t_model)
+                elif sda_stategy == "earliest_models":
+                    if len(teacher_models) < sda_teachers:
+                        t_model = copy.deepcopy(model)
+                        t_model.eval()
+                        teacher_models.append(t_model)
+                elif sda_stategy == 'latest_model':
+                    t_model = copy.deepcopy(model)
+                    t_model.eval()
+                    teacher_models = [t_model]
+                elif sda_stategy == 'clone_models':
+                    if len(teacher_models) == 0:
+                        t_model = copy.deepcopy(model)
+                    else:
+                        t_model = copy.deepcopy(teacher_models[-1])
+                    t_model.eval()
+                    teacher_models.append(t_model)
+                    if len(teacher_models) > sda_teachers:
+                        teacher_models = teacher_models[1:sda_teachers + 1]
+
+                #  if len(teacher_models) < sda_teachers:
+                #      if teacher_models:
+                #          tmodel = copy.deepcopy(teacher_models[-1])
+                #      else:
+                #          t_model = copy.deepcopy(model)
+                #      t_model.eval()
+                #      teacher_models.append(t_model)
 
             print(" ")
             if 'cuda' in str(args.device):
@@ -488,13 +721,16 @@ class Trainer:
         return trained_steps, train_loss / trained_steps
 
     def evaluate(self, args, model, eval_examples):
-        eval_dataset, eval_features = self.examples_to_dataset(
-            eval_examples, args.eval_max_seq_length)
+        #  _, eval_features = self.examples_to_dataset(eval_examples,
+        #                                              args.eval_max_seq_length)
+        eval_features = self.encode_examples(eval_examples,
+                                             args.eval_max_seq_length)
 
-        self.on_eval_start(args, eval_dataset)
+        self.on_eval_start(args, eval_features)
         eval_dataloader = generate_dataloader(
             args,
-            eval_dataset,
+            #  eval_dataset,
+            eval_features,
             batch_size=args.per_gpu_eval_batch_size,
             keep_order=True,
             collate_fn=self.collate_fn)
@@ -515,16 +751,19 @@ class Trainer:
                        desc=f"Evaluating")
         for step, batch in enumerate(eval_dataloader):
             model.eval()
-            batch = tuple(t.to(args.device) for t in batch)
+            batch = tuple(
+                t.to(args.device) if isinstance(t, torch.Tensor) else t
+                for t in batch)
             with torch.no_grad():
                 #  inputs = self.batch_to_inputs(args, batch)
                 #  outputs = model(**inputs)
-                batch_features = eval_features[args.per_gpu_eval_batch_size *
-                                               step:args.
-                                               per_gpu_eval_batch_size *
-                                               (step + 1)]
-                outputs, results = self.on_eval_step(args, model, step, batch,
-                                                     batch_features)
+                outputs, results = self.on_eval_step(args, model, step, batch)
+                #  batch_features = eval_features[args.per_gpu_eval_batch_size *
+                #                                 step:args.
+                #                                 per_gpu_eval_batch_size *
+                #                                 (step + 1)]
+                #  outputs, results = self.on_eval_step(args, model, step, batch,
+                #                                       batch_features)
 
                 # -------- loss --------
                 loss = outputs[0]
@@ -543,7 +782,7 @@ class Trainer:
             values += [('loss', loss.item())]
             pbar.update(step + 1, values=values)
 
-        results = self.on_eval_end(args, eval_dataset)
+        results = self.on_eval_end(args, eval_features)
 
         self.eval_loss = eval_loss / eval_steps
         results['loss'] = self.eval_loss
@@ -557,11 +796,13 @@ class Trainer:
         if is_multi_processes(args) and not pred_output_dir.exists():
             os.makedirs(pred_output_dir)
 
-        test_dataset, _ = self.examples_to_dataset(test_examples,
-                                                   args.eval_max_seq_length)
+        #  test_dataset, _ = self.examples_to_dataset(test_examples,
+        #                                             args.eval_max_seq_length)
+        test_features = self.encode_examples(test_examples,
+                                             args.eval_max_seq_length)
         test_dataloader = generate_dataloader(
             args,
-            test_dataset,
+            test_features,
             batch_size=args.per_gpu_predict_batch_size,
             keep_order=True,
             collate_fn=self.collate_fn)
@@ -572,10 +813,12 @@ class Trainer:
         logger.info(f"  Batch size = {args.per_gpu_predict_batch_size}")
 
         pbar = Progbar(target=len(test_dataloader), desc=f"Predicting")
-        self.on_predict_start(args, test_dataset)
+        self.on_predict_start(args, test_features)
         for step, batch in enumerate(test_dataloader):
             model.eval()
-            batch = tuple(t.to(args.device) for t in batch)
+            batch = tuple(
+                t.to(args.device) if isinstance(t, torch.Tensor) else t
+                for t in batch)
             with torch.no_grad():
                 #  inputs = self.package_inputs_from_batch(args,
                 #                                          batch,
@@ -587,5 +830,5 @@ class Trainer:
 
             pbar.update(step + 1)
 
-        results = self.on_predict_end(args, test_dataset)
+        results = self.on_predict_end(args, test_features)
         return results
